//
//The little touches that don't directly relate to Mario's moveset
//

local MUSIC_FORCERESET = 16384

//Mario skins
if not MarioSkins
	rawset(_G, "MarioSkins", {})
end
MarioSkins["mario"] = true
MarioSkins["luigi"] = true
MarioSkins["toad"] = true
MarioSkins["peach"] = true
MarioSkins["daisy"] = true
MarioSkins["wario"] = true
MarioSkins["waluigi"] = true
MarioSkins["bowser"] = true
MarioSkins["bowserjr"] = true
MarioSkins["yoshi"] = true
MarioSkins["birdo"] = true
MarioSkins["goomba"] = true
MarioSkins["koopa"] = true
MarioSkins["piranhaplant"] = true
MarioSkins["cappy"] = true
MarioSkins["rosalina"] = true
MarioSkins["luma"] = true
MarioSkins["sgimario"] = true

//Constants - Should be the same as in mario.lua
//Yes this is a bad coding practice, but shut up!

local JUMPSTATE = S_MARIO_JUMP1
local FALLSTATE = S_MARIO_FALL

local DOUBLEJUMPSTATE = S_MARIO_DOUBLE
local DOUBLEFALLSTATE = S_MARIO_DOUBLEFALL

local TRIPLEJUMPSTATE = S_MARIO_TRIPLE

local WALLJUMPSTATE = DOUBLEJUMPSTATE

local CROUCHSTATE = S_PLAY_GLIDE_LANDING
local CRAWLSTATE = S_PLAY_CLIMB

local SPINSTATE = S_MARIO_SPIN
local DIVESTATE = S_PLAY_SPINDASH

local GROUNDPOUNDENDSTATE = S_MARIO_GROUNDPOUND
local GROUNDPOUNDSTARTSTATE = S_MARIO_POUNDSTART

local BALLOONSTATE = S_MARIO_BALLOON
local CAPESPINSTATE = S_MARIO_CAPESPIN //cape.lua

local LAVASTATE = S_MARIO_LAVA

local SQUISHFACTOR = 32

local NIGHTSTRANS = freeslot("S_MARIO_NIGHTS1")
states[S_MARIO_NIGHTS1] = {SPR_PLAY, SPR2_TAL3|FF_ANIMATE, 18, A_ForceStop, 5,  3, freeslot("S_MARIO_NIGHTS2")}
states[S_MARIO_NIGHTS2] = {SPR_PLAY, SPR2_TAL4, 23, A_ForceStop, 0,  0, S_PLAY_NIGHTS_FLOAT}

//Mario mode stomp damage for characters without stomp damage
addHook("PlayerCanDamage", function(player, mobj)
	local mo = player.mo
	
	if PSO or not (mariomode and not (player.charflags & SF_STOMPDAMAGE)
	and mo and mo.valid and not P_IsObjectOnGround(mo)
	and not P_PlayerInPain(player) and mobj and mobj.valid) then return end
	
	local bottomheight = mo.z
	local topheight = mo.z + mo.height

	if (mo.eflags & MFE_VERTICALFLIP)
		local swap = bottomheight
		bottomheight = topheight
		topheight = swap
	end

	if P_MobjFlip(mo)*(bottomheight - mobj.z + mobj.height/2) > 0
		return true
	end
end)

//Special Sound Effects
local mariosounds = {
	[sfx_shldls] = sfx_mario8,
	[sfx_shield] = sfx_mario3,
	[sfx_attrsg] = sfx_mrmini,
	[sfx_forcsg] = sfx_mario3,
	[sfx_armasg] = sfx_mrpbln,
	[sfx_wirlsg] = sfx_mrcape,
	[sfx_elemsg] = sfx_mrmetl,
	[sfx_s3k3e] = sfx_mario3,
	[sfx_s3k3f] = sfx_mario3,
	[sfx_s3k41] = sfx_mrglpw,
	[sfx_bkpoof] = sfx_mrblow,
	//[sfx_itemup] = sfx_mario4,
	[sfx_strpst] = sfx_mrchek,
	[sfx_token] = sfx_mrtoke,
	[sfx_cgot] = sfx_mremer,
	
	//Amy
	[sfx_s3k42] = sfx_mrhmr3,
	[sfx_s3k8b] = sfx_mrhmr4,
}

/*if sfx_rnglft and sfx_rngrgh
	mariosounds[sfx_rnglft] = sfx_mario4
	mariosounds[sfx_rngrgh] = sfx_mario4
end*/

for oldsound, newsound in pairs(mariosounds)
	sfxinfo[newsound].priority = sfxinfo[oldsound].priority
end

local function StopMultipleSounds(mo, sound, amount)
	if not sound then return end
	for i = 0, amount - 1
		S_StopSoundByID(mo, sound + i)
	end
end

addHook("PostThinkFrame", do for player in players.iterate
	local mo = player.mo
	if not (mo and mo.valid) then continue end
	
	player.mariowassliding = player.pflags & PF_SLIDING
	
	//Cancel stinkiness if we're not smelly anymore
	if mo.mariostinky
		if player.panim != PA_IDLE or mo.skin != "mario"
			mo.mariostinky = 0
			if not mo.stinkymario
				P_RestoreMusic(player)
			end
		end
	elseif mo.stinkymario and not (mo.stinkymario.valid
	and mo.stinkymario.mariostinky
	and R_PointToDist2(0, mo.z, R_PointToDist2(mo.x, mo.y, mo.stinkymario.x, mo.stinkymario.y), mo.stinkymario.z) < 32*mo.radius)
		mo.stinkymario = nil
		P_RestoreMusic(player)
	end
	
	//Look, we're forcing the powerup sizes whether you like it or not!
	if player.mariosize
		mo.spritexscale = FRACUNIT + player.mariosize
		mo.spriteyscale = mo.spritexscale
	end
	
	//Squash players
	if player.mariosquashed
		local squish = 0
		player.mariosquashed = $ - 1
		
		if player.mariosquashstart < TICRATE
			player.mariosquashstart = $ + 4
			squish = FRACUNIT - ease.outback(player.mariosquashstart*FRACUNIT/TICRATE)
		else
			if player.pflags & PF_SPINDOWN
			and P_IsObjectOnGround(mo) and not player.speed
				player.mariosquashed = min($, TICRATE)
			end
			squish = ease.inoutback(FRACUNIT - player.mariosquashed*FRACUNIT/TICRATE)
		end
		
		if player.mariosquashed == TICRATE
			S_StartSound(mo, sfx_mrestr)
		end
		
		if not player.mariosize or not player.mariosquashed
			mo.spriteyscale = FRACUNIT
			mo.spritexscale = FRACUNIT
		end
		
		if player.mariosquashed
			mo.spriteyscale = $/2 + squish/2
			mo.spritexscale = $*3/2 - squish/2
		end
	end
	
	//Not a Mario skin!
	if not MarioSkins[mo.skin] and not mariomode
		player.marioprevscoreadd = player.scoreadd
		continue
	end
	
	//Sound replacements
	for oldsound, newsound in pairs(mariosounds)
		if S_SoundPlaying(mo, oldsound)
			S_StopSoundByID(mo, oldsound)
			S_StartSound(mo, newsound)
		end
	end
	
	//Super Ball sound for Amy's shield
	if IsMario(mo) and S_SoundPlaying(mo, sfx_mario3)
	and player.powers[pw_shield] == SH_SUPERBALL
		S_StopSoundByID(mo, sfx_mario3)
		S_StartSound(mo, sfx_mrball)
	end
	
	//1up combo breaker (not while mega)
	if not MarioScoreGametype()
	and player.marioprevscoreadd != nil
	and player.marioprevscoreadd != player.scoreadd
	and (not player.mariomega or player.powers[pw_invulnerability])
		if player.scoreadd > 8 and not mariomode and not modeattacking
			P_GivePlayerLives(player, 1)
			S_StartSound(mo, sfx_marioa)
			
			local score = P_SpawnMobjFromMobj(mo, 0,0,0, MT_SCORE)
			if score and score.valid
				score.state = S_SCRK
			end
		end
	end
	player.marioprevscoreadd = player.scoreadd
	
	//Bonking your head
	if player.marioprevmomz == nil
		//nothing lol
	elseif not mariomode and not P_IsObjectOnGround(mo)
	and mo.health and not player.powers[pw_carry]
	and (mo.z == mo.floorz or mo.z + mo.height == mo.ceilingz)
	and mo.momz*P_MobjFlip(mo) <= 0 and player.marioprevmomz*P_MobjFlip(mo) > 0
		//Bonk sound
		if not mo.mariohithead
			S_StartSound(mo, sfx_mario1)
			mo.mariohithead = 2
		end
		
		//Break bustables when jumping into them from below
		if MarioSkins[mo.skin] and mo.subsector and mo.subsector.valid
		and mo.subsector.sector and mo.subsector.sector.valid
			for fof in mo.subsector.sector.ffloors()
				if not (fof.flags & FF_EXISTS) continue end -- Does it exist?
				if not (fof.flags & FF_BUSTUP) continue end -- Is it bustable?
				
				if mo.z + mo.momz + mo.height < fof.bottomheight continue end -- Are we too low?
				if mo.z + mo.momz > fof.topheight continue end -- Are we too high?  
				
				EV_CrumbleChain(fof) -- Crumble
				if fof.master.flags & ML_EFFECT5
					P_LinedefExecute(P_AproxDistance(fof.master.dx, fof.master.dy)>>FRACBITS, mo, fof.target)
				end
				
				S_StopSound(fof.target)
				S_StartSound(fof.target, sfx_mrbrck)
			end
		end
	elseif mo.mariohithead
		mo.mariohithead = $ - 1
	end
	
	//Mario's continuous sfx, moved here from the PlayerThink to not be a frame late
	if not IsMario(mo) then continue end
	
	//Force Mario and Luigi's stats for persona
	if srb2p and MARIO_STATS[mo.skin]
		if player.runspeed != MARIO_STATS[mo.skin].runspeed
			for field, value in pairs(MARIO_STATS[mo.skin])
				player[field] = value
			end
		end
	end
	
	//Force state for Luigi
	if mo.luigiability and player.powers[pw_carry] == CR_PLAYER
	and mo.tracer and mo.tracer.player
		player.drawangle = mo.tracer.player.drawangle
	end
	
	//Force the squishy
	if player.powers[pw_shield] == SH_BALLOON
		if mo.marioexplode
			if not player.mariosize
				mo.spritexscale = FRACUNIT
			end
			mo.spritexscale = $ + FixedMul($,
			sin(FixedAngle(mo.marioexplode*16*FRACUNIT))/4
			+ mo.marioexplode*UINT8_MAX*3)
			mo.spriteyscale = mo.spritexscale
		elseif player.mariosquish
			local factor = player.mariosquish*FRACUNIT/SQUISHFACTOR
			if not player.mariosize
				mo.spritexscale = FRACUNIT
				mo.spriteyscale = FRACUNIT
			end
			mo.spritexscale = $ + FixedMul($, factor)
			mo.spriteyscale = $ - FixedMul($, factor)
		end
	end
	
	//Undo skid sfx
	if (player.marioprevstate == S_PLAY_SKID or player.marioskid)
	and mo.state != S_PLAY_SKID
		S_StopSoundByID(mo, skins[player.skin].soundsid[SKSSKID])
		player.marioskid = nil
	end
	
	//SRB2 Infinity stuff
	if PSO and player.pso
		//Wall jump dodge roll
		if PSO.lib.playerInputDown(player, IPT_INTERACT) == 1
		and player.mariowallslide
		and not nododge
		and not player.pso.action
		and player.normalspeed
		and not (player.powers[pw_nocontrol] 
		or player.pso.menu.ref 
		or player.pso.palette.open 
		or P_PlayerInPain(player) 
		or player.pso.teleport 
		or player.pso.jumppad 
		or player.pso.dodgeroll
		or player.pso.stagger
		or player.pso.blowdown
		or player.pso.knockback
		or player.pso.condition == COND_FREEZE)
		
			if player.pso.pp < PSO.info.dodgeroll_cost
				PSO.lib.noPPNotif(player)
				return
			end
			
			player.pso.pp = $ - PSO.info.dodgeroll_cost
			player.pso.dodgeroll = TICRATE/3
			if PSO.lib.hasSkill(player, "all_dodgeextension")
				player.pso.dodgeroll = $*3/2
			end
			
			P_InstaThrust(mo, player.drawangle, mo.scale*20)
			mo.momz = mo.scale*7
			
			MarioCircle(mo, MT_MINECARTSPARK, 6, 4*mo.scale, FixedDiv(mo.height, mo.scale)*2/3,
			mo.scale/4, 2, 0, true, player.drawangle)
			
			mo.state = WALLJUMPSTATE
			S_StopSoundByID(mo, sfx_mrsld1)
			S_StopSoundByID(mo, sfx_mrsld2)
			S_StartSound(mo, sfx_mrwljm)
			
			MarioPlayVoice(player, "jump")
			player.mariopsowalljump = true
		end
		
		if player.mariopsowalljump
			if player.pso.dodgeroll and mo.momz*P_MobjFlip(mo) > 0
				mo.state = WALLJUMPSTATE
			else
				player.mariopsowalljump = false
			end
		end
		
		//Special dodge roll animations
		if player.pso.dodgeroll
		and player.marioprevstate != S_PLAY_ROLL and mo.state == S_PLAY_ROLL
			local pmx = player.rmomx + player.cmomx
			local pmy = player.rmomy + player.cmomy
			local diff = R_PointToAngle2(0,0, pmx, pmy) - player.drawangle
			
			//Side flip
			if player.marioprevstate == S_PLAY_SKID
				mo.state = S_MARIO_SIDEFLIP
				mo.state = S_MARIO_SIDEFLIP
				mo.state = S_MARIO_SIDEFLIP
				S_StartSound(mo, sfx_mrflip)
				
				MarioPlayVoice(player, "spin")
				
			//Triple jump
			else
				if player.mariotriple < 3
					player.mariotriple = $+1
				else
					player.mariotriple = 1
				end
				player.mariotripletime = 5
				
				if player.mariotriple <= 1
					mo.state = S_PLAY_JUMP
					MarioPlayVoice(player, "jump")
				elseif player.mariotriple == 2
					mo.state = DOUBLEJUMPSTATE
					MarioPlayVoice(player, "double")
				else
					mo.state = TRIPLEJUMPSTATE
					MarioPlayVoice(player, "triple")
				end
				
				player.drawangle = R_PointToAngle2(0,0, mo.momx, mo.momy)
				player.mariojumpdir = player.drawangle
			end
		end
		
		//Voice clips for strong attacks
		if player.mariopsotimer != player.pso.actiontimer
			local action = (player.pso.taction or player.pso.action)
			action = $ and PSO.info.photonArts[$]
			
			
			if action
				//Strong attack voice
				if not action.combo
				and ((player.mariopsogrounded and not P_IsObjectOnGround(mo))
				or (player.skidtime == 20 and player.pso.actionspeed == 0)
				or (player.pso.actiontimer == 2
				and action.casttime and action.casttime >= TICRATE/2))
					MarioPlayVoice(player, "attack")
				end
				player.mariotriple = 0
			end
			player.mariopsotimer = player.pso.actiontimer
		end
		player.mariopsogrounded = P_IsObjectOnGround(mo)
	end
	
	//Init the state stuff
	if not mo.marioprevstateset
		mo.marioprevstateset = true
		player.marioprevstate = mo.state
	
	//P Balloon
	elseif player.powers[pw_shield] == SH_BALLOON and mo.health
		if not player.mariopain and mo.state != BALLOONSTATE
			if mo.state == S_PLAY_RIDE
				mo.state = BALLOONSTATE
				mo.sprite2 = SPR2_TAL6
			else
				mo.state = BALLOONSTATE
			end
		end
		player.marioprevstate = mo.state
	
	//Shoe!!
	elseif player.marioshoe and player.marioshoe.valid
		if not player.mariopain and mo.health
		and mo.state != SPINSTATE and mo.state != CAPESPINSTATE
			mo.state = S_TAILSOVERLAY_0DEGREES
			mo.tics = -1
			player.panim = PA_DASH
		end
		
		player.marioshoe.tracer = mo
		A_CapeChase(player.marioshoe, 1,0)
		
		player.marioprevstate = mo.state
		
	//Nights transformation
	elseif mo.state == S_PLAY_NIGHTS_TRANS1
	and player.powers[pw_carry] == CR_NIGHTSMODE
		mo.state = NIGHTSTRANS
		A_ForceStop(mo)
		
		mo.reactiontime = mo.tics + states[states[NIGHTSTRANS].nextstate].tics
		player.marioprevstate = mo.state
	
	//Don't bother with the dive and spin states
	elseif mo.state == DIVESTATE or mo.state == SPINSTATE or MarioCantMove(player)
		//sorry nothing
		
	//Hammer sprung
	elseif mo.eflags & MFE_SPRUNG and mo.state == S_PLAY_SPRING
	and (player.marioprevstate == S_PLAY_MELEE
	or player.marioprevstate == S_PLAY_MELEE_FINISH
	or player.marioprevstate == S_PLAY_MELEE_LANDING)
		mo.momx = $*5/4
		mo.momy = $*5/4
		mo.momz = $*5/4

		S_StartSound(mo, sfx_sprong)
		DoMarioHammerHit(player, true)
		mo.state = S_PLAY_SPRING
	
	//Normal jump fall state
	elseif player.marioprevstate == S_PLAY_JUMP and mo.state == S_PLAY_FALL
		mo.state = FALLSTATE
		
	//Double jump fall state
	elseif (mo.state == DOUBLEJUMPSTATE or mo.state == WALLJUMPSTATE)
	and mo.momz*P_MobjFlip(mo) <= 0 and player.panim == PA_JUMP
		mo.state = DOUBLEFALLSTATE
		
	//Fire speed
	elseif mo.state == S_PLAY_FIRE_FINISH
		mo.tics = min($, 10)
		
	//Fire fall state
	elseif not P_IsObjectOnGround(mo)
	and player.marioprevstate == S_PLAY_FIRE_FINISH and mo.state == S_PLAY_STND
		mo.state = FALLSTATE
		
	//Cape fall state
	elseif not P_IsObjectOnGround(mo)
	and player.marioprevstate == CAPESPINSTATE and mo.state == S_PLAY_WALK
		mo.state = S_PLAY_FALL
		
	//Force crouch frames
	elseif player.mariocrouch and player.panim and player.panim <= PA_DASH
	and mo.state != CROUCHSTATE and mo.state != CRAWLSTATE
		if player.skidtime
			player.skidtime = 0
			S_StopSoundByID(mo, skins[player.skin].soundsid[SKSSKID])
		end
		
		mo.state = CROUCHSTATE
		mo.tics = 2
		player.panim = PA_DASH
		
	//Skip the ground pound start's last frame
	elseif mo.state == GROUNDPOUNDSTARTSTATE and mo.frame & FF_FRAMEMASK >= D
		mo.state = GROUNDPOUNDENDSTATE
		mo.tics = -1
	
	//Edge voice clip
	elseif player.marioprevstate != S_PLAY_EDGE
	and states[player.marioprevstate].nextstate != S_PLAY_EDGE
	and mo.state == S_PLAY_EDGE
		MarioPlayVoice(player, "edge")
		
	//Face the direction we jump in
	//Do the jump transition state too
	elseif ((player.marioprevstate != S_PLAY_JUMP and mo.state == S_PLAY_JUMP)
	or (player.marioprevstate != DOUBLEJUMPSTATE and mo.state == DOUBLEJUMPSTATE)
	or (player.marioprevstate != TRIPLEJUMPSTATE and mo.state == TRIPLEJUMPSTATE))
		player.marioprevstate = mo.state
		mo.state = JUMPSTATE
		
		if player.marioprevstate == TRIPLEJUMPSTATE
			mo.sprite2 = SPR2_SWIM
		end
		
		if (player.cmd.forwardmove or player.cmd.sidemove)
		and not player.mariowalljump and not player.mariowallslide
		and not (player.pflags & (PF_JUMPSTASIS))
		and player.pflags & (PF_JUMPED|PF_STARTJUMP|PF_DIRECTIONCHAR)
			player.drawangle = MarioInputAngle(player)
			player.mariojumpdir = nil
		end
	end
	
	//Return JUMPSTATE to what it was before
	if mo.state == states[JUMPSTATE].nextstate and player.marioprevstate
		mo.state = player.marioprevstate
		if player.marioprevstate == S_PLAY_JUMP
		or player.marioprevstate == DOUBLEJUMPSTATE
			mo.state = player.marioprevstate
			player.panim = PA_FALL
			mo.tics = 5
		end
		
		//Run jump
		if (player.speed >= FixedMul(player.runspeed, mo.scale) or player.mariocapeflight)
		and (mo.state == S_PLAY_JUMP or mo.state == DOUBLEJUMPSTATE)
			mo.state = S_TAILSOVERLAY_FLY
		end
	//Get the prevstate
	elseif mo.state != JUMPSTATE and not player.mariopain
		player.marioprevstate = mo.state
	end
	
	//Slide sound
	if player.marioslidesfx
		if player.marioslidesfx == 1
			StopMultipleSounds(mo, sfx_mrsld1, 2)
			S_StartSound(mo, sfx_mrsld1)
		elseif player.marioslidesfx > 15 and not ((player.marioslidesfx - 10) % 12)
			StopMultipleSounds(mo, sfx_mrsld1, 2)
			S_StartSound(mo, sfx_mrsld2)
		end
		
		player.marioslidesfx = $ + 1
		player.mariowasslidesfx = true
	elseif player.mariowasslidesfx
		StopMultipleSounds(mo, sfx_mrsld1, 2)
		player.mariowasslidesfx = false
	end
	
	//Spin sound
	if player.mariospinsfx
		if player.mariospinsfx == 1
			StopMultipleSounds(mo, sfx_mrspn1, 4)
			S_StartSound(mo, sfx_mrspn1)
		elseif player.mariospinsfx > 15 and not ((player.mariospinsfx - 10) % 12)
			StopMultipleSounds(mo, sfx_mrspn1, 4)
			S_StartSound(mo, sfx_mrspn2)
		elseif player.mariospinsfx == -1
			StopMultipleSounds(mo, sfx_mrspn1, 4)
			S_StartSound(mo, sfx_mrspn3)
		elseif player.mariospinsfx < -15 and not ((player.mariospinsfx + 10) % 12)
			StopMultipleSounds(mo, sfx_mrspn1, 4)
			S_StartSound(mo, sfx_mrspn4)
		end
		
		if player.mariospinsfx > 0
			player.mariospinsfx = $ + 1
		else
			player.mariospinsfx = $ - 1
		end
		
		if not player.mariospinjump
			player.mariospinsfx = 0
		end
		player.mariowasspinsfx = true
	elseif player.mariowasspinsfx
		StopMultipleSounds(mo, sfx_mrspn1, 4)
		player.mariowasspinsfx = false
	end
	
	//Gold shine sound
	if srb2p //nope
	elseif player.powers[pw_shield] == SH_GOLDFLOWER
	and not player.mariopain
		if not player.mariogoldsound
		or player.mariogoldsound >= 88
			player.mariogoldsound = 0
			S_StartSound(mo, sfx_mrgold)
		end
		player.mariogoldsound = $ + 1
	else
		if player.mariogoldsound and S_SoundPlaying(mo, sfx_mrgold)
			S_StopSoundByID(mo, sfx_mrgold)
		end
		player.mariogoldsound = 0
	end
	
	//Do overalls thinker
	DoMarioOveralls(player, mo, player.followmobj)
	
	//Do cape thinker
	DoMarioCape(player, mo)
end end)

//Coins everywhere!! Though only in singleplayer
addHook("MobjSpawn", function(mobj)
	if not mariomode and ((not multiplayer
	and consoleplayer and consoleplayer.valid
	and MarioSkins[skins[consoleplayer.skin].name])
	or mariocoins.value)
		mobj.type = MT_COIN
		mobj.state = mobj.info.spawnstate
	end
end, MT_RING)
addHook("MobjSpawn", function(mobj)
	if not mariomode and ((not multiplayer
	and consoleplayer and consoleplayer.valid
	and MarioSkins[skins[consoleplayer.skin].name])
	or mariocoins.value)
		mobj.type = MT_FLINGCOIN
		mobj.state = mobj.info.spawnstate
	end
end, MT_FLINGRING)

//Coin shadows
local function CoinShadow(mo)
	if not mo.shadowscale
		mo.shadowscale = 2*FRACUNIT/3
	end
end
addHook("MobjSpawn", CoinShadow, MT_COIN)
addHook("MobjSpawn", CoinShadow, MT_FLINGCOIN)

//Mario always loses coins, not rings!
addHook("TouchSpecial", function(mobj, mo)
	if not mariomode and mobj and mobj.valid and mobj.target == mo
	and mo and mo.valid and mo.player and MarioSkins[mo.skin]
	and not P_CanPickupItem(mo.player, false)
		mobj.type = MT_FLINGCOIN
		mobj.state = mobj.info.spawnstate
	end
end, MT_FLINGRING)

//Nights trail effect
/*
local trailcolors = {
	SKINCOLOR_RED,
	SKINCOLOR_ORANGE,
	SKINCOLOR_SUPERGOLD5,
	SKINCOLOR_PERIDOT,
	SKINCOLOR_EMERALD,
	SKINCOLOR_AQUA,
	SKINCOLOR_SKY,
	SKINCOLOR_BLUE,
	SKINCOLOR_PURPLE,
	SKINCOLOR_MAGENTA,
}
addHook("MobjThinker", function(mobj)
	if mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS and not mobj.redstar
	and mobj.target and mobj.target.valid and IsMario(mobj.target)
		mobj.redstar = true
		
		if mobj.target.state == S_PLAY_NIGHTS_STAND
			P_RemoveMobj(mobj)
			return
		end
		
		mobj.state = S_OBJPLACE_DUMMY
		mobj.tics = 161
		mobj.sprite = SPR_FLAM
		mobj.frame = M
		
		mobj.scale = $/5
		mobj.scalespeed = mobj.scale/mobj.tics
		mobj.destscale = $/8
		mobj.spriteyoffset = 32*FRACUNIT
		
		mobj.color = mobj.target.color
		mobj.colorized = true
		mobj.renderflags = $ | RF_FULLBRIGHT
		mobj.blendmode = AST_ADD
		
		if mobj.target.player and mobj.target.player.powers[pw_nights_superloop]
			mobj.color = trailcolors[mobj.target.player.powers[pw_nights_superloop]/2 % #trailcolors + 1]
		end
	end
end, MT_NIGHTSPARKLE)
*/

//Define all the music we're gonna replace
local mariojingles = {
	["_inv"] = "_minv",
	["_shoes"] = "mmega",
	["_endb"] = "mendb",
	["_endg"] = "mendg",
	["_clear"] = "mclear",
	["_1up"] = "m1up",
	["_clea2"] = "mclea2",
	["_drown"] = "mdrown",
	["_gover"] = "mgover",
	["_inter"] = "minter",
	["_super"] = "bk2mad",
	//["_creds"] = "mcreds", //Don't think 1MB is worth it
	["_conti"] = "mconti",
}

//Now for the Super Ball stuff...gbs files sure are convienent
local mariolandmusic = {
	//Jingles
	["_inv"] = 10,
	["_minv"] = 10,
	["_shoes"] = 13,
	["_clear"] = 12,
	["_1up"] = 15,
	["_clea2"] = -12,
	["_drown"] = 17,
	["_gover"] = 18,
	["_inter"] = -13,
	["mrmenu"] = -13,
	["_conti"] = -10,
	["mrdead"] = 11,
	["disco"] = 6,
	["_creds"] = 6,
	["hystnk"] = -23,
	
	//Forest
	["gfzalt"] = -20,
	["mp_for"] = -20,
	["mp_fr2"] = -20,
	
	//Beach
	["mp_coa"] = 3,
	["mp_ghz"] = 3,
	["mp_sun"] = 3,
	
	//Machinery (the proper song doesn't work :[ )
	["thz1"] = -20,
	["thz2"] = -20,
	["thzalt"] = -20,
	
	//Ruins
	["dsz1"] = 2,
	["dsz2"] = 2,
	["atz"] = 2,
	["mp_wtr"] = 2,
	["mp_wt2"] = 2,
	["mp_glz"] = 2,
	["mp_gl2"] = 2,
	["mario2"] = 2,
	
	//Haunted
	["hauntd"] = -17,
	["mp_des"] = -17,
	["mp_min"] = -17,
	["dcz"] = -17,
	
	//Haunted House
	["cnz1pr"] = -31,
	
	//Wario's Castle
	["cez1"] = -11,
	["cez2"] = -11,
	["mp_rai"] = -11,
	
	//Panic
	["cezr"] = -35,
	["bpnch1"] = -33,
	["bpnch2"] = -35,
	
	//Athletic
	["acz1"] = -1,
	["acz2"] = -1,
	["agz"] = -1,
	["agzalt"] = -1,
	["mp_air"] = -1,
	
	//Eastern
	["rvz1"] = 5,
	["rvz2"] = 5,
	["rvzalt"] = 5,
	["mp_fir"] = 5,
	
	//Space Zone
	["erz1"] = -19,
	["mp_dom"] = -19,
	
	//Star Maze
	["erz2"] = -29,
	["mp_0ri"] = -29,
	["mp_mrz"] = -29,
	
	//Special
	["spec1"] = 4,
	["spec2"] = 4,
	["spec3"] = 4,
	["spec4"] = 4,
	["spec5"] = 4,
	["spec6"] = 4,
	["spec7"] = 4,
	["spec8"] = 4,
	["ospec"] = 4,
	["ospec2"] = 4,
	
	//Bosses
	["vsboss"] = -32,
	["vsfang"] = -32,
	["vsalt"] = -32,
	
	//Final boss
	["bcz1"] = -33,
	["vsmetl"] = -38,
	["bcz2"] = -14,
	["vsbrak"] = 9,
}

//Let's try and prevent crashes for SRB2P by hijacking the music functions directly
if srb2p
	//Add in the songs to our table
	for i = 1, 6
		mariojingles["block"..i] = "mrblk"..i
	end
	for i = 1, 3
		mariojingles["reslt"..i] = "mrwin"..i
	end
	for i = 1, 5
		mariojingles["vrch"..i] = "mrvrc"..i
	end
	mariojingles["vrst"] = "mrvrc0"
	mariojingles["shop"] = "mrshop"
	mariojingles["peak"] = "mrpeak"
	
	//Allow forcing Mario music in Persona when not Mario
	rawset(_G, "mariomusic", CV_RegisterVar({
		name = "mariomusic",
		defaultvalue = 0,
		flags = CV_NETVAR|CV_CALL,
		PossibleValue = CV_OnOff,
		func = function(cv)
			if cv.value
				print("Mario's music will now be forced for all skins.")
			else
				print("Mario's music will now be exclusive to Mario and Luigi.")
			end
		end
	}))
	
	//Dungeon music
	local Mario_getTartarusMusic = DNG_getTartarusMusic
	DNG_getTartarusMusic = function(block)
		local music = Mario_getTartarusMusic(block)
		local jingle = mariojingles[string.lower(music)]
		
		if consoleplayer and consoleplayer.valid
		and (MarioSkins[skins[consoleplayer.skin].name] or mariomode or mariomusic.value)
		and jingle and S_MusicExists(jingle)
			return jingle
		else
			return music
		end
	end
	
	//Void Run music
	local Mario_setFloor = DNG_setFloor
	DNG_setFloor = function(f)
		local oldmusic = S_MusicName()
		local oldmusicposition = S_GetMusicPosition()
		
		Mario_setFloor(f)
		local newmusic = S_MusicName()
		local jingle = mariojingles[string.lower(newmusic)]
		
		//This is a hack, but oh well
		if gamemap != srb2p.tartarus_play and not server.cdungeon
		and server.gamemode == GM_VOIDRUN
		and consoleplayer and consoleplayer.valid
		and (MarioSkins[skins[consoleplayer.skin].name] or mariomode or mariomusic.value)
		and jingle and S_MusicExists(jingle)
			COM_BufInsertText(consoleplayer, "tunes "..jingle)
			if string.lower(oldmusic) == jingle
				S_SetMusicPosition(oldmusicposition)
			end
		end
	end
	
	//Results music
	local Mario_PlayRandomBattleMusic = MUS_PlayRandomBattleMusic
	MUS_PlayRandomBattleMusic = function(tab)
		local music = Mario_PlayRandomBattleMusic(tab)
		local jingle = mariojingles[string.lower(music)]
		
		if consoleplayer and consoleplayer.valid
		and (MarioSkins[skins[consoleplayer.skin].name] or mariomode or mariomusic.value)
		and jingle and S_MusicExists(jingle)
			return jingle
		else
			return music
		end
	end
	
	//Shop 1 (The Shop)
	local Mario_startShop = DNG_startShop
	DNG_startShop = function(mo, shop)
		Mario_startShop(mo, shop)
		if server.gamemode ~= GM_VOIDRUN and S_MusicExists("MRSHOP")
		and (MarioSkins[mo.skin] or mariomode or mariomusic.value)
			S_ChangeMusic("MRSHOP", true, mo.player, nil, nil, 400)
		end
	end
	
	//Shop 2 (Equip Lab)
	local Mario_startEquipLab = DNG_startEquipLab
	DNG_startEquipLab = function(mo)
		Mario_startEquipLab(mo)
		if server.gamemode ~= GM_VOIDRUN and S_MusicExists("MRSHOP")
		and (MarioSkins[mo.skin] or mariomode or mariomusic.value)
			S_ChangeMusic("MRSHOP", true, mo.player, nil, nil, 400)
		end
	end
	
	//Shop 3 (Void Run)
	local Mario_getVoidRunRewards = D_getVoidRunRewards
	D_getVoidRunRewards = function()
		Mario_getVoidRunRewards()
		if server.P_DungeonStatus.VR_challenge%3 == 0 and S_MusicExists("MRSHOP")
		and consoleplayer and consoleplayer.valid
		and (MarioSkins[skins[consoleplayer.skin].name] or mariomode or mariomusic.value)
			S_ChangeMusic("MRSHOP", true, consoleplayer, nil, nil, 400)
		end
	end
	
	//The Peak
	addHook("MapLoad", do
		if mapheaderinfo[gamemap].musname == "PEAK"
		and consoleplayer and consoleplayer.valid
		and (MarioSkins[skins[consoleplayer.skin].name] or mariomode or mariomusic.value)
		and S_MusicExists("MRPEAK")
			S_ChangeMusic("MRPEAK", true, consoleplayer, nil, nil, 400)
		end
	end)
	
	return //Don't add the MusicChange hook!
end

//Not in SRB2P, let's replace some musics!
addHook("MusicChange", function(oldname, newname, mflags, looping)
	oldname = string.lower($)
	newname = string.lower($)
	
	for i, player in ipairs({consoleplayer, secondarydisplayplayer})
		if not (player and player.valid) then continue end
		
		if oldname == "luwins" and (gamestate == GS_INTERMISSION
		or (player.mo and player.mo.luiginumberone))
			return true
		end
		
		local superball = IsMario(player) and not player.powers[pw_super]
		and (player.powers[pw_shield] == SH_SUPERBALL
		or (gamestate == GS_INTERMISSION and player.marioshield == SH_SUPERBALL))
		if newname == "mland1"
			newname = player.playerstate and "mrdead" or oldname
			superball = true
		end
		
		if newname == "_shoes" and not superball
		and (not IsMario(player) or not player.mariomega)
			continue
		end
		
		if (MarioSkins[skins[player.skin].name] or mariomode)
		and (mariojingles[newname] or superball)
			if newname == "_clear"
			and mapheaderinfo[gamemap].levelflags & LF_WARNINGTITLE
				newname = "_clea2"
			end
			local newjingle = superball and mariolandmusic[newname] or mariojingles[newname]
			if newjingle == "mendg"
				if skins[player.skin].name == "luigi"
					newjingle = "mendl"
				elseif not IsMario(player)
					newjingle = "mendb"
				end
			end
			if newjingle and type(newjingle) != "number" and not S_MusicExists(newjingle) then return end
			
			if newjingle == nil and superball
				return "mland1", MUSIC_FORCERESET, true
			elseif type(newjingle) == "number"
				if newjingle >= 0
					newjingle = $ - 1
					return "mland1", newjingle|MUSIC_FORCERESET, true
				else
					newjingle = abs($) - 1
					return "mland2", newjingle|MUSIC_FORCERESET, true
				end
			elseif newname == "_inv" or newname == "_shoes"
				return newjingle, nil, true
			else
				return newjingle
			end
		end
	end
end)